using UnityEngine;
using System.Collections;

public class Game : MonoBehaviour {
	
	public enum State
	{
		INITIAL,
		SHOWING_WELCOME_MESSAGE,
		MOVING,
		SLIDING,
		SLIDING_ADJUSTING,
		SHOWING_LEVEL_WON_MESSAGE,
		GAME_COMPLETE,
		SHOWING_GAME_OVER_MESSAGE,
		COUNT
	}
	
	private class SectionInfo
	{
		public SectionType sectionType = SectionType.I;
		// 0, 1, 2 or 3
		public int rotation = 0;
		public bool visited = false;
		public GameObject sectionObject = null;
		
		public SectionInfo(SectionType sectionType, int rotation)
		{
			this.sectionType = sectionType;
			this.rotation = rotation;
		}
	
		public bool IsNorthAccessible()
		{
			return GetAccessibilityString().Contains("N");
		}
		
		public bool IsEastAccessible()
		{
			return GetAccessibilityString().Contains("E");
		}
		
		public bool IsSouthAccessible()
		{
			return GetAccessibilityString().Contains("S");
		}
		
		public bool IsWestAccessible()
		{
			return GetAccessibilityString().Contains("W");
		}
		
		//TODO make private again
		public string GetAccessibilityString()
		{
			string accessibilityString = "";
			
			if (sectionType == SectionType.I) // deemed OK
			{
				if (rotation == 0 || rotation == 2) // OK for 2
				{
					/*					  
					  X   X
					  X   X
					  X   X					  
					 */					
					accessibilityString += "NS";
				}
				else if (rotation == 1 || rotation == 3) // OK for 1
				{
					/*					  
					  X X X
					       
					  X X X					  
					 */					
					accessibilityString += "EW";
				}
			}
			else if (sectionType == SectionType.L) // deemed OK
			{
				if (rotation == 0)
				{
					/*					  
					  X   X
					      X
					  X X X					  
					 */					
					accessibilityString += "NW"; //OK
				}
				else if (rotation == 1)
				{
					/*					  
					  X   X
					  X   
					  X X X					  
					 */					
					accessibilityString += "NE"; //OK
				}
				else if (rotation == 2)
				{
					/*					  
					  X X X
					  X   
					  X   X					  
					 */					
					accessibilityString += "ES"; 
				}
				else if (rotation == 3)
				{
					/*					  
					  X X X
					      X
					  X   X					  
					 */					
					accessibilityString += "SW"; 
				}	
			}
			else if (sectionType == SectionType.T) // deemed OK
			{
				if (rotation == 0)
				{
					/*					  
					  X   X
					      X
					  X   X					  
					 */					
					accessibilityString += "NSW"; 
				}
				else if (rotation == 1) //OK
				{
					/*					  
					  X   X
					      
					  X X X					  
					 */					
					accessibilityString += "NEW"; 
				}
				else if (rotation == 2) // OK
				{
					/*					  
					  X   X
					  X   
					  X   X					  
					 */					
					accessibilityString += "NES"; 
				}
				else if (rotation == 3)
				{
					/*					  
					  X X X
					      
					  X   X					  
					 */					
					accessibilityString += "ESW"; 
				}				
			}
			else if (sectionType == SectionType.X) //OK
			{
				/*					  
				  X   X
				      
				  X   X					  
				 */					
				accessibilityString += "NESW"; //OK
			}
			else if (sectionType == SectionType.S) // deemed OK
			{
				if (rotation == 0)
				{
					/*					  
					  X X X
					  X   X
					  X   X					  
					 */					
					accessibilityString += "S"; 
				}
				else if (rotation == 1)
				{
					/*					  
					  X X X
					      X
					  X X X					  
					 */					
					accessibilityString += "W"; 
				}
				else if (rotation == 2)
				{
					/*					  
					  X   X
					  X   X
					  X X X					  
					 */					
					accessibilityString += "N"; //OK
				}
				else if (rotation == 3)
				{
					/*					  
					  X X X
					  X   
					  X X X					  
					 */					
					accessibilityString += "E"; //OK
				}
			}
			return accessibilityString;
		}
	}
	
	private enum SectionType
	{
		I,
		L,
		T,
		X,
		S,
		COUNT,
	}
	
	// TUNING
	private const float SLIDE_BUTTON_OFFSET_X = 150.0f;
	private const float SLIDE_BUTTON_OFFSET_Y = 150.0f;
	private const float SLIDE_BUTTON_SIZE_X = 75.0f;
	private const float SLIDE_BUTTON_SIZE_Y = 75.0f;
	private const int MIN_LEVEL_SIZE = 2;
	private const int MAX_LEVEL_SIZE = 5; // to allow bigger levels, the collision boxes should probably become bigger
	private const int HEALTH_DECREASE_SLIDE = 5;
	public const int HEALTH_DECREASE_WALK = 1;
	private const int SCORE_INCREASE_FULL_EXPLORATION = 150;
	private const int SCORE_INCREASE_MONEY = 25;
	private const int SCORE_INCREASE_TREASURE = 50;
	private Rect CONTINUE_BUTTON_RECT;
	private Rect PUSH_BUTTON_RECT;
	private Rect MESSAGE_BOX_TEXT_RECT;
	
	private static int innerLevelSize = MIN_LEVEL_SIZE;	
	private GameObject[] sections = null;	
	private GameObject floor = null;
	private Player player = null;
	private LevelObject target = null;
	private ArrayList moneyObjects = null;
	private ArrayList medikitObjects = null;
	private ArrayList treasureObjects = null;
	private GameObject moneyPrefab = null;
	private GameObject medikitPrefab = null;
	private GameObject treasurePrefab = null;
	private SectionInfo[,] maze = null;
	private Vector2 targetPosition = Vector2.zero;
	private Rect SLIDE_LEFT_BUTTON_RECT;
	private Rect SLIDE_RIGHT_BUTTON_RECT;
	private Rect SLIDE_BACKWARD_BUTTON_RECT;
	private static State state = State.INITIAL;
	private int score = 0;
	public static int health = 100;
	private Texture hudTex_scoreHealthBg = null;
	private Texture hudTex_scoreHealthText = null;
	private Texture hudTex_messageBox = null;
	private Texture hudTex_explored = null;
	private Texture hudTex_minimapBg = null;
	private Texture hudTex_minimapPlayerN = null;
	private Texture hudTex_minimapPlayerE = null;
	private Texture hudTex_minimapPlayerS = null;
	private Texture hudTex_minimapPlayerW = null;
	private Texture hudTex_minimapFixed = null;
	private Texture hudTex_minimapFixedVisited = null;
	private Texture hudTex_minimapRegular = null;
	private Texture hudTex_minimapRegularVisited = null;
	private Texture hudTex_minimapExit = null;
	private Texture hudTex_controllerPush = null;
	
	private AudioClip sfxGameCompleted;
	private AudioClip sfxGameOver;
	private AudioClip sfxLevelCompleted;
	private AudioClip sfxMedikit;
	private AudioClip sfxMoney;
	private AudioClip sfxSliding1;
	private AudioClip sfxSliding2;
	private AudioClip sfxSliding3;
	private AudioClip sfxTreasure;
	private AudioClip level1Music;
	private AudioClip level2Music;
	private AudioClip level3Music;

	// State.MOVING
	public static bool ignoringInput = false;
	
	// State.SLIDING
	private SectionInfo slidingSectionInfo = null;
	private GameObject slidingSectionObject = null;
	private GameObject slidingSectionObjectCopy = null;
	private Vector3 slidingSectionObjectCopyStartPosition;
	private Vector3 slidingSectionObjectCopyEndPosition;
	private float slidingTimeLeft = 0.0f;
	private Vector3 slidingPlayerStartPosition;
	private Vector3 slidingPlayerEndPosition;
	private const float SLIDING_TIME = 7.5f / 4.0f; //TUNING
	
	// State.SLIDING_ADJUSTING
	private Vector3[,] adjustingStartPositions;
	private Quaternion[,] adjustingStartQuaternions;
	private float adjustingTimeLeft = 0.0f;
	private const float ADJUSTING_TIME = 0.75f; //TUNING NOTE: 0.50 too short, the old section hasn't fallen off screen yet
	
	private Quaternion[] sectionRotationQuaternions = new Quaternion[4]
	{
		// 0
		Quaternion.AngleAxis(90.0f * 0.0f, Vector3.up),
		// 1
		Quaternion.AngleAxis(90.0f * 1.0f, Vector3.up),
		// 2
		Quaternion.AngleAxis(90.0f * 2.0f, Vector3.up),
		// 3
		Quaternion.AngleAxis(90.0f * 3.0f, Vector3.up),
	};
	
	public void Awake()
	{
		//TODO remove		
		//Physics.gravity = new Vector3(-20.0f, -10.0f, 0.0f);
		
		// TUNING
		CONTINUE_BUTTON_RECT = new Rect((Screen.width - 100.0f)/2.0f - 10.0f, Screen.height - 135.0f + 25.0f, 140.0f, 60.0f);
		PUSH_BUTTON_RECT = new Rect(265.0f, 310.0f, 66.0f, 60.0f);
		MESSAGE_BOX_TEXT_RECT = new Rect(125.0f, 50.0f - 50.0f, Screen.width - 250.0f, Screen.height - 100.0f);
		
		// get gameobjects
		floor = GameObject.Find("Floor");
		player = (Player)GameObject.Find("Player").GetComponent(typeof(LevelObject));
		target = (LevelObject)GameObject.Find("Target").GetComponent(typeof(LevelObject));
		moneyPrefab = (GameObject)Resources.Load("Prefabs/Money");
		medikitPrefab = (GameObject)Resources.Load("Prefabs/Medikit");
		treasurePrefab = (GameObject)Resources.Load("Prefabs/Treasure");
		
		// get textures
		hudTex_scoreHealthBg = (Texture)Resources.Load("Textures/score-health_bg");
		hudTex_scoreHealthText = (Texture)Resources.Load("Textures/score-health");
		hudTex_messageBox = (Texture)Resources.Load("Textures/messagebox");
		hudTex_explored = (Texture)Resources.Load("Textures/explored");
		hudTex_minimapBg = (Texture)Resources.Load("Textures/minimap_bg");
		hudTex_minimapPlayerN = (Texture)Resources.Load("Textures/playerN");
		hudTex_minimapPlayerE = (Texture)Resources.Load("Textures/playerE");
		hudTex_minimapPlayerS = (Texture)Resources.Load("Textures/playerS");
		hudTex_minimapPlayerW = (Texture)Resources.Load("Textures/playerW");
		hudTex_minimapFixed = (Texture)Resources.Load("Textures/fixed");
		hudTex_minimapFixedVisited = (Texture)Resources.Load("Textures/fixed_visited");
		hudTex_minimapRegular = (Texture)Resources.Load("Textures/regular");
		hudTex_minimapRegularVisited = (Texture)Resources.Load("Textures/regular_visited");
		hudTex_minimapExit = (Texture)Resources.Load("Textures/exit");
		hudTex_controllerPush = (Texture)Resources.Load("Textures/controller_push");
		
		sfxGameCompleted = (AudioClip)Resources.Load("Sfx/game_completed");
		sfxGameOver = (AudioClip)Resources.Load("Sfx/game_over");
		sfxLevelCompleted = (AudioClip)Resources.Load("Sfx/level_completed");
		sfxMedikit = (AudioClip)Resources.Load("Sfx/medikit");
		sfxMoney = (AudioClip)Resources.Load("Sfx/money");
		sfxSliding1 = (AudioClip)Resources.Load("Sfx/sliding1");
		sfxSliding2 = (AudioClip)Resources.Load("Sfx/sliding2");
		sfxSliding3 = (AudioClip)Resources.Load("Sfx/sliding3");
		sfxTreasure = (AudioClip)Resources.Load("Sfx/treasure");		
		level1Music = (AudioClip)Resources.Load("Music/level1");
		level2Music = (AudioClip)Resources.Load("Music/level2");
		level3Music = (AudioClip)Resources.Load("Music/level3");
		
		sections = new GameObject[(int)SectionType.COUNT]
		{
			//I
			(GameObject)Resources.Load("Prefabs/Section_I"),
			//L
			(GameObject)Resources.Load("Prefabs/Section_L"),
			//T
			(GameObject)Resources.Load("Prefabs/Section_T"),
			//X
			(GameObject)Resources.Load("Prefabs/Section_X"),
			//S
			(GameObject)Resources.Load("Prefabs/Section_S"),
		};
		
		// ignore collision between SectionBoxColliders and LevelObjects
		Physics.IgnoreLayerCollision(8, 9);
		
		Reset();
	}
	
	public void Start()
	{
		
	}
	
	public void LateUpdate()
	{
		ignoringInput = false;
	}
	
	public void Update()
	{		
		if (state == State.MOVING)
		{
			Vector2 playerSectionIndices = player.GetSectionIndices();
			
			player.SetMoveForwardPossible(IsMoveForwardPossible());
			player.SetMoveBackwardPossible(IsMoveBackwardPossible());
			
			maze[(int)playerSectionIndices.x, (int)playerSectionIndices.y].visited = true;
			
			if (playerSectionIndices == targetPosition)
			{
				if (innerLevelSize == MAX_LEVEL_SIZE)
				{
					ChangeState(State.GAME_COMPLETE);
					audio.Stop();
					audio.PlayOneShot(sfxGameCompleted);
				}
				else
				{				
					ChangeState(State.SHOWING_LEVEL_WON_MESSAGE);
					audio.Stop();
					audio.PlayOneShot(sfxLevelCompleted);
				}
			}
			else if (health <= 0)
			{
				ChangeState(State.SHOWING_GAME_OVER_MESSAGE);
				audio.Stop();
				audio.PlayOneShot(sfxGameOver);
			}
			else
			{
				bool objectFound = false;
				foreach(LevelObject moneyObject in moneyObjects)
				{
					if (moneyObject.GetSectionIndices() == playerSectionIndices)
					{
						audio.PlayOneShot(sfxMoney);
						Destroy(moneyObject.gameObject);
						moneyObjects.Remove(moneyObject);
						score += SCORE_INCREASE_MONEY;
						break;
					}
				}
				
				if (!objectFound)
				{
					foreach(LevelObject treasureObject in treasureObjects)
					{
						if (treasureObject.GetSectionIndices() == playerSectionIndices)
						{
							audio.PlayOneShot(sfxTreasure);
							Destroy(treasureObject.gameObject);
							treasureObjects.Remove(treasureObject);
							score += SCORE_INCREASE_TREASURE;
							break;
						}
					}
						
					if (!objectFound)
					{
						foreach(LevelObject medikitObject in medikitObjects)
						{
							if (medikitObject.GetSectionIndices() == playerSectionIndices)
							{
								audio.PlayOneShot(sfxMedikit);
								Destroy(medikitObject.gameObject);
								medikitObjects.Remove(medikitObject);
								health = 100;
								break;
							}
						}
					}
				}
			}
		}
		else if (state == State.SLIDING)
		{		
			float  t = 1.0f - (slidingTimeLeft/SLIDING_TIME);
			
			slidingSectionObjectCopy.transform.position = MathExt.Lerp( slidingSectionObjectCopyStartPosition,
				                                                      slidingSectionObjectCopyEndPosition,
				                                                      t );
			
			player.transform.position = MathExt.Lerp( slidingPlayerStartPosition,
				  	   						  		  slidingPlayerEndPosition,
													  t );
			
			slidingTimeLeft -= Time.deltaTime;
			if (slidingTimeLeft < 0.0f)
			{
				ChangeState(State.SLIDING_ADJUSTING);
			}
		}
		else if (state == State.SLIDING_ADJUSTING)
		{
			// interpolation of position and quaternion for all sections
			float t = 1.0f - (adjustingTimeLeft / ADJUSTING_TIME);
			for (int x=0; x<GetOuterLevelSize(); ++x)
			{
				for (int z=0; z<GetOuterLevelSize(); ++z)
				{
					maze[x,z].sectionObject.transform.position = MathExt.Lerp( adjustingStartPositions[x,z],
																			   SectionIndexToCoordinate(new Vector2(x,z)),
																			   t );
					maze[x,z].sectionObject.transform.rotation = Quaternion.Lerp( adjustingStartQuaternions[x,z],
																			      sectionRotationQuaternions[(int)maze[x,z].rotation],
																			      t );
				}
			}

			adjustingTimeLeft -= Time.deltaTime;
			if (adjustingTimeLeft < 0.0f)
			{
				ChangeState(State.MOVING);
			}
		}
	}
	
	public void OnGUI()	
	{
		// set all labels color
		GUI.skin.label.normal.textColor = Color.black;	
		
		if (state == State.MOVING)
		{			
			if (IsSlidingForwardPossible())
			{
				GUI.DrawTexture(PUSH_BUTTON_RECT, hudTex_controllerPush);	
				
				if ((PUSH_BUTTON_RECT.Contains(Util.GetInvertedMousePosition()) && Input.GetMouseButtonDown(0)) || Input.GetKeyDown(KeyCode.Space))
				{
					if (!ignoringInput)
					{
						StartSlideForward();
						health -= HEALTH_DECREASE_SLIDE;
					}
				}
			}
			
			OnGUI_HUD();
		}
		else if (state == State.SLIDING)
		{
			OnGUI_HUD();
		}
		else if (state == State.SLIDING_ADJUSTING)
		{
			OnGUI_HUD();
		}
		else if (state == State.SHOWING_LEVEL_WON_MESSAGE)
		{		

			// create level end message
			string message = "Congratulations!\nYou will now continue to level " + innerLevelSize + "!\n\n"  +
						     "Score: " + score + "\n";
			int displayScore = score;
			{
				int percentageExplored = (int)Mathf.Abs(GetPercentageExplored());
				if (percentageExplored == 100)
				{
					message += "Full exploration bonus: " + SCORE_INCREASE_FULL_EXPLORATION + "\n";
					displayScore += SCORE_INCREASE_FULL_EXPLORATION;
				}
				else
				{
					message += "Exploration bonus: " + percentageExplored + "\n";
					displayScore += percentageExplored;
				}			
								 
				message += "Health regenerated: " + GetLevelEndHealthIncrement();
				message += "\n\nNew score: " + displayScore; 
			}
	
			if (ShowMessageBox(message))
			{
				health += GetLevelEndHealthIncrement();
				
				int percentageExplored = (int)Mathf.Abs(GetPercentageExplored());
				if (percentageExplored == 100)
				{
					score += SCORE_INCREASE_FULL_EXPLORATION;
				}
				else
				{
					score += percentageExplored;
				}				
				
				++innerLevelSize;
				Reset();
			}
		}
		else if (state == State.GAME_COMPLETE)
		{
			// create end game message
			string message = "Congratulations!\nYou completed the game.\n\n";
			{
				message += "Score: " + score + "\n";
				
				int displayScore = score;
				
				int percentageExplored = (int)Mathf.Abs(GetPercentageExplored());
				if (percentageExplored == 100)
				{
					message += "Full exploration bonus: " + SCORE_INCREASE_FULL_EXPLORATION + "\n";
					displayScore += SCORE_INCREASE_FULL_EXPLORATION;
				}
				else
				{
					message += "Exploration bonus: " + percentageExplored + "\n";
					displayScore += percentageExplored;
				}
				
				message += "Health remaining: " + Mathf.Max(health, 0.0f) + "\n\n";
				displayScore += health;
				
				message += "FINAL SCORE: " + displayScore;
			}
			
			if (ShowMessageBox(message))
			{
				ResetGame();
			}
		}
		else if (state == State.SHOWING_GAME_OVER_MESSAGE)
		{
			// create game over message
			string message = "You ran out of health\n\n";
			{
				message += "FINAL SCORE: " + score + "\n\n\nPlease try again!";
			}
			
			if (ShowMessageBox(message))
			{
				ResetGame();
			}
		}
		else if (state == State.SHOWING_WELCOME_MESSAGE)
		{
			//TODO write welcome message
			string message = "-[ Scarab Amazing ]-\n\n\n" +
					         "You just woke up in the top floor of a pyramid.\n\nTry to find as much presents as possible, and then take the door to the next floor.\n " +
					         "\nNavigate trough the maze by using the on-screen buttons or your keyboard's arrow buttons and space bar.\n" +
					         "If you're stuck, you can move the maze and reach new places. Be careful though, this takes a lot of your powers! If your health runs out before reaching the exit, you're in trouble!\n" +
					         "\nYou cannot move the maze anywhere: the circles on the map cannot be moved.\n\nGood luck!";
			
			if (ShowMessageBox(message))
			{
				ChangeState(State.MOVING);
			}
		}
	}
	
	void OnGUI_HUD()
	{
		// score and health
		GUI.DrawTexture(new Rect(Screen.width - (512.0f/2.0f) + 60.0f, -30.0f + 25.0f, 512.0f / 2.0f, 257.0f / 2.0f), hudTex_scoreHealthBg);
		GUI.DrawTexture(new Rect(Screen.width - (288.0f / 2.0f) - 50.0f + 50.0f, 25.0f - 15.0f + 10.0f, 288.0f / 2.0f, 68.0f / 2.0f), hudTex_scoreHealthText);
		FontWriter.WriteString(new Rect(440.0f, 55.0f, 60.0f, 0.0f), ""+ Mathf.Max(0, score));
		FontWriter.WriteString(new Rect(530.0f, 55.0f, 60.0f, 0.0f), "" + Mathf.Max(0, health));	
		
		// minimap
		GUI.DrawTexture(new Rect(-40.0f - 40.0f, -20.0f, 512.0f / 2.0f + 10.0f, 385.0f / 2.0f + 10.0f), hudTex_minimapBg);
		
		Rect miniMapRect = new Rect(5.0f, 5.0f, 145.0f, 145.0f);
		Vector2 miniMapSpacing = new Vector2(miniMapRect.width / GetOuterLevelSize(), miniMapRect.height / GetOuterLevelSize());
		
		Rect currentRect = new Rect(3.0f, 3.0f, miniMapSpacing.x * 1.5f, miniMapSpacing.y * 1.5f);
		for (int y=GetOuterLevelSize()-1; y>=0; --y)
		{
			for (int x=0; x<GetOuterLevelSize(); ++x)
			{
				if ((x == (int)player.GetSectionIndices().x) && (y == (int)player.GetSectionIndices().y))
				{
					//TODO make sure the player is rotated depending on its orientation (see GetArrowCharacter)
					string direction = player.GetDirection();
					if (direction == "N")
					{
						GUI.DrawTexture(currentRect, hudTex_minimapPlayerN);
					}
					else if (direction == "E")
					{
						GUI.DrawTexture(currentRect, hudTex_minimapPlayerE);
					}
					else if (direction == "S")
					{
						GUI.DrawTexture(currentRect, hudTex_minimapPlayerS);
					}
					else if (direction == "W")
					{
						GUI.DrawTexture(currentRect, hudTex_minimapPlayerW);
					}
				}
				else if ((x == targetPosition.x) && (y == targetPosition.y))
				{
					GUI.DrawTexture(currentRect, hudTex_minimapExit);
				}
				else if (maze[x,y].visited)
				{
					if (IsMoveablePosition(x,y))
					{
						GUI.DrawTexture(currentRect, hudTex_minimapRegularVisited);
					}
					else
					{
						GUI.DrawTexture(currentRect, hudTex_minimapFixedVisited);
					}
				}
				else
				{
					if (IsMoveablePosition(x,y))
					{
						GUI.DrawTexture(currentRect, hudTex_minimapRegular);
					}
					else
					{
						GUI.DrawTexture(currentRect, hudTex_minimapFixed);
					}
				}
				
				currentRect.x += miniMapSpacing.x;
			}
			currentRect.x = 0.0f;
			currentRect.y += miniMapSpacing.y;
		}

		
		// display discovery percentage	
		GUI.DrawTexture(new Rect(Screen.width - (288.0f / 2.0f) - 50.0f + 30.0f, 25.0f - 15.0f + 10.0f + 60.0f, 138.0f / 2.0f, 66.0f / 2.0f), hudTex_explored);	
		string percentageString = "" + (int)GetPercentageExplored() + "%";
		FontWriter.WriteString(new Rect(520.0f, 85.0f, 60.0f, 0.0f), percentageString);
					
		// display compass
		/*
		string arrowCharacter = GetArrowCharacter();
		string compass =   "/      N      \\\n"
			             + "\n"
			             + "W    " + arrowCharacter + "     E\n"
				         + "\n"
				         + "\\      S      /\n";
		GUI.Label(new Rect(Screen.width - 100.0f, 250.0f, Screen.width, Screen.height), compass);	
		*/				
	}
	
	public static int GetOuterLevelSize()
	{
		return InnerCoordinateToOuter(innerLevelSize);
	}
	
	public static State GetState()
	{
		return state;
	}
	
	private void ResetGame()
	{
		innerLevelSize = MIN_LEVEL_SIZE;
		score = 0;
		health = 100;
		Reset();
	}
	
	private void Reset()
	{
		state = State.INITIAL;
		//ChangeState(State.MOVING);
		
		// NOTE make sure welcome message isn't shown every level!
		ChangeState(State.SHOWING_WELCOME_MESSAGE);
		
		// delete any existing sections
		foreach (Section section in Object.FindObjectsOfType(typeof(Section)))
		{
			Destroy(section.gameObject);
		}
		
		// delete existing money objects, medikits and treasures
		foreach (Money money in Object.FindObjectsOfType(typeof(Money)))
		{
			Destroy(money.gameObject);
		}
		foreach (Medikit medikit in Object.FindObjectsOfType(typeof(Medikit)))
		{
			Destroy(medikit.gameObject);
		}	
		foreach (Treasure treasure in Object.FindObjectsOfType(typeof(Treasure)))
		{
			Destroy(treasure.gameObject);
		}
		
		floor.transform.localScale = new Vector3(GetOuterLevelSize() / 10.0f, 1.0f, GetOuterLevelSize() / 10.0f);
		
		// determine random maze
		maze = new SectionInfo[GetOuterLevelSize(), GetOuterLevelSize()];
		for (int x=0; x<GetOuterLevelSize(); ++x)
		{
			for (int z=0; z<GetOuterLevelSize(); ++z)
			{				
				maze[x,z] = GetRandomSectionInfo();
			}
		}

		//NOTE these positions won't be the same as the player can't start on unmoveable positions, while items have to
		RandomizePlayerPosition();
		RandomizeTargetPosition();
		
		CreateOtherLevelObjects();
		
		// improve random maze
		{
			// make sure target cannot be reached (is surrounded by walls)
			{
				SectionInfo targetSectionInfo = maze[(int)targetPosition.x, (int)targetPosition.y];
				
				// N of target
				if (targetSectionInfo.IsNorthAccessible() && ((int)targetPosition.y < GetOuterLevelSize() - 1))
				{
					while (maze[(int)targetPosition.x, (int)targetPosition.y + 1].IsSouthAccessible())
					{
						maze[(int)targetPosition.x, (int)targetPosition.y + 1] = GetRandomSectionInfo();
					}
				}
				
				// E of target
				if (targetSectionInfo.IsEastAccessible() && ((int)targetPosition.x < GetOuterLevelSize() - 1))
				{
					while (maze[(int)targetPosition.x + 1, (int)targetPosition.y].IsWestAccessible())
					{
						maze[(int)targetPosition.x + 1, (int)targetPosition.y] = GetRandomSectionInfo();
					}
				}
				
				// S of target
				if (targetSectionInfo.IsSouthAccessible() && ((int)targetPosition.y > 0))
				{
					while (maze[(int)targetPosition.x, (int)targetPosition.y - 1].IsNorthAccessible())
					{
						maze[(int)targetPosition.x, (int)targetPosition.y - 1] = GetRandomSectionInfo();
					}
				}
				
				// W of target
				if (targetSectionInfo.IsWestAccessible() && ((int)targetPosition.x > 0))
				{
					while (maze[(int)targetPosition.x - 1, (int)targetPosition.y].IsEastAccessible())
					{
						maze[(int)targetPosition.x - 1, (int)targetPosition.y] = GetRandomSectionInfo();
					}					
				}
			}
		}

		// instantiate maze
		for (int x=0; x<GetOuterLevelSize(); ++x)
		{
			for (int z=0; z<GetOuterLevelSize(); ++z)
			{
				GameObject sectionObject = CreateSectionObject(maze[x, z], SectionIndexToCoordinate(new Vector2(x,z)));
				
				sectionObject.transform.name += "(" + x + ", " + z + ")";
				((Section)sectionObject.GetComponent(typeof(Section))).AllowMove(IsMoveablePosition(x,z));
				maze[x, z].sectionObject = sectionObject;
			}
		}
		
		// play music
		audio.Stop();
		int musicIndex = Mathf.Min(2, innerLevelSize - MIN_LEVEL_SIZE);
		if (musicIndex == 0)
		{
			audio.clip = level1Music;
		}
		else if (musicIndex == 1)
		{
			audio.clip = level2Music;
		}
		else if (musicIndex == 2)
		{
			audio.clip = level3Music;
		}
		audio.loop = true;
		audio.Play();
	}
	
	private Vector2 GetRandomTargetPosition()
	{
		int randomInnerX = Random.Range(0, innerLevelSize);
		int randomInnerZ = Random.Range(0, innerLevelSize);
		return new Vector2( InnerCoordinateToOuter(randomInnerX),
			                InnerCoordinateToOuter(randomInnerZ) );
	}
	
	private Vector2 GetRandomPlayerPosition()
	{
		return new Vector2( Random.Range(0, GetOuterLevelSize()),
			                Random.Range(0, GetOuterLevelSize()) );
	}
	
	private static int InnerCoordinateToOuter(int innerCoordinate)
	{
		/*
		    example
		    
		 	0,0 -> 1,1
		 	0,1 -> 1,3
		 	2,1 -> 5,3
		*/
		
		return (innerCoordinate * 2) + 1;
	}
	
	private void RandomizePlayerPosition()
	{
		Vector2 randomPosition = GetRandomPlayerPosition();	
		// from these positions, the player cannot move, so can get stuck
		while (!IsMoveablePosition((int)randomPosition.x, (int)randomPosition.y))
		{
			randomPosition = GetRandomPlayerPosition();
		}
		
		player.transform.position = new Vector3( randomPosition.x - (GetOuterLevelSize()/2.0f) + 0.5f,
			                                     player.transform.position.y,
			                                     randomPosition.y  - (GetOuterLevelSize()/2.0f) + 0.5f );
	}
	
	private void RandomizeTargetPosition()
	{
		targetPosition = GetRandomTargetPosition();
		target.transform.position = new Vector3( targetPosition.x - (GetOuterLevelSize()/2.0f) + 0.5f,
			                                     0.0f,
			                                     targetPosition.y  - (GetOuterLevelSize()/2.0f) + 0.5f );				
	}
	
	private bool IsMoveablePosition(int x, int z)
	{
		return ((x % 2 == 0) || (z % 2 == 0));
	}
	
	private SectionInfo GetRandomSectionInfo()
	{
		return new SectionInfo( (SectionType)Random.Range(0, (int)SectionType.COUNT),
						        Random.Range(0, 4) );
	}
	
	private bool IsMoveForwardPossible()
	{		
		return CanLeaveSectionForward() && CanEnterSectionForward() && !IsFacingEdge();
	}
	
	private bool IsMoveBackwardPossible()
	{		
		return CanLeaveSectionBackward() && CanEnterSectionBackward() && !IsBackingEdge();
	}
	
	private bool CanLeaveSectionForward()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool canLeaveSectionForward = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			canLeaveSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y].IsNorthAccessible();
		}
		else if (playerDirection == "E")
		{
			canLeaveSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y].IsEastAccessible();
		}
		else if (playerDirection == "S")
		{
			canLeaveSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y].IsSouthAccessible();
		}
		else if (playerDirection == "W")
		{
			canLeaveSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y].IsWestAccessible();
		}
		
		return canLeaveSectionForward;
	}
	
	private bool CanEnterSectionForward()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool canEnterSectionForward = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			if (playerIndices.y < GetOuterLevelSize()-1)
			{
				canEnterSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y+1].IsSouthAccessible();
			}
		}
		else if (playerDirection == "E")
		{ 
			if (playerIndices.x < GetOuterLevelSize()-1)
			{
				canEnterSectionForward = maze[(int)playerIndices.x+1,(int)playerIndices.y].IsWestAccessible();
			}
		}
		else if (playerDirection == "S")
		{
			if (playerIndices.y > 0)
			{
				canEnterSectionForward = maze[(int)playerIndices.x,(int)playerIndices.y-1].IsNorthAccessible();
			}
		}
		else if (playerDirection == "W")
		{
			if (playerIndices.x > 0)
			{
				canEnterSectionForward = maze[(int)playerIndices.x-1,(int)playerIndices.y].IsEastAccessible();
			}
		}
		
		return canEnterSectionForward;
	}
	
	private bool CanLeaveSectionBackward()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool canLeaveSectionBackward = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			canLeaveSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y].IsSouthAccessible();
		}
		else if (playerDirection == "E")
		{
			canLeaveSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y].IsWestAccessible();
		}
		else if (playerDirection == "S")
		{
			canLeaveSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y].IsNorthAccessible();
		}
		else if (playerDirection == "W")
		{
			canLeaveSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y].IsEastAccessible();
		}
		
		return canLeaveSectionBackward;
	}
	
	private bool CanEnterSectionBackward()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool canEnterSectionBackward = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			if (playerIndices.y > 0)
			{
				canEnterSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y-1].IsNorthAccessible();
			}
		}
		else if (playerDirection == "E")
		{
			if (playerIndices.x > 0)
			{
				canEnterSectionBackward = maze[(int)playerIndices.x-1,(int)playerIndices.y].IsEastAccessible();
			}
		}
		else if (playerDirection == "S")
		{
			if (playerIndices.y < GetOuterLevelSize()-1)
			{
				canEnterSectionBackward = maze[(int)playerIndices.x,(int)playerIndices.y+1].IsSouthAccessible();
			}
		}
		else if (playerDirection == "W")
		{
			if (playerIndices.x < GetOuterLevelSize()-1)
			{
				canEnterSectionBackward = maze[(int)playerIndices.x+1,(int)playerIndices.y].IsWestAccessible();
			}
		}
		
		return canEnterSectionBackward;
	}
	
	private void StartSlideLeft()
	{
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y , -1);
		}
		else if (playerDirection == "E")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, 1);
		}
		else if (playerDirection == "S")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, 1);
		}
		else if (playerDirection == "W")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, -1);
		}		
	}
	
	private void StartSlideRight()
	{
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, 1);
		}
		else if (playerDirection == "E")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, -1);
		}
		else if (playerDirection == "S")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, -1);
		}
		else if (playerDirection == "W")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, 1);
		}
	}
	
	private void StartSlideForward()
	{
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, 1);
		}
		else if (playerDirection == "E")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, 1);
		}
		else if (playerDirection == "S")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, -1);
		}
		else if (playerDirection == "W")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, -1);
		}
	}
	
	private void StartSlideBackward()
	{
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, -1);
		}
		else if (playerDirection == "E")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, -1);
		}
		else if (playerDirection == "S")
		{
			StartSlideZAxis((int)player.GetSectionIndices().x, 1);
		}
		else if (playerDirection == "W")
		{
			StartSlideXAxis((int)player.GetSectionIndices().y, 1);
		}
	}
	
	private void StartSlideXAxis(int z, int direction)
	{
		PlayRandomSlideSound();
		
		if (direction < 0) // left
		{
			// slidingSectionInfo: leftmost section the row
			slidingSectionInfo = maze[0, z];
			
			// copy gameObject
			Vector3 sectionCopyPosition = slidingSectionInfo.sectionObject.transform.position;
			sectionCopyPosition.x += GetOuterLevelSize();
			slidingSectionObjectCopy = CreateSectionObject(slidingSectionInfo, sectionCopyPosition);
			
			// set other variables required for sliding
			slidingSectionObjectCopyStartPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition = sectionCopyPosition;
			slidingPlayerStartPosition = player.transform.position;
			slidingPlayerEndPosition = player.transform.position;
			if (player.GetSectionIndices().y == z)
			{
				slidingPlayerEndPosition.x += direction;
			}

			slidingSectionObjectCopyEndPosition.x -= 1.0f;
			
			// modify maze
			for (int x=1; x<GetOuterLevelSize(); ++x)
			{
				maze[x-1, z] = maze[x, z];
			}
			maze[GetOuterLevelSize()-1, z] = slidingSectionInfo;
			
			// remember the section object that will be dropped
			slidingSectionObject = slidingSectionInfo.sectionObject;
			
			// set the new section object in the maze
			slidingSectionInfo.sectionObject = slidingSectionObjectCopy;
		}
		else // right
		{
			// slidingSectionInfo: rightmost section of the row
			slidingSectionInfo = maze[GetOuterLevelSize()-1 ,z];
			
			// copy gameObject
			Vector3 sectionCopyPosition = slidingSectionInfo.sectionObject.transform.position;
			sectionCopyPosition.x -= GetOuterLevelSize();
			slidingSectionObjectCopy = CreateSectionObject(slidingSectionInfo, sectionCopyPosition);
			
			// set other variables required for sliding
			slidingSectionObjectCopyStartPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition.x += 1.0f;
			slidingPlayerStartPosition = player.transform.position;
			slidingPlayerEndPosition = player.transform.position;
			if (player.GetSectionIndices().y == z)
			{
				slidingPlayerEndPosition.x += direction;
			}
			
			// modify maze
			for (int x=GetOuterLevelSize()-2; x>=0; --x)
			{
				maze[x+1, z] = maze[x, z];
			}
			maze[0, z] = slidingSectionInfo;

			// remember the section object that will be dropped
			slidingSectionObject = slidingSectionInfo.sectionObject;
			
			// set the new section object in the maze
			slidingSectionInfo.sectionObject = slidingSectionObjectCopy;
		}
		
		ChangeState(State.SLIDING);
	}
	
	private void StartSlideZAxis(int x, int direction)
	{
		PlayRandomSlideSound();
		
		if (direction > 0) // forward
		{
			// slidingSectionInfo: frontmost section of the column
			slidingSectionInfo = maze[x, GetOuterLevelSize()-1];
			
			// copy gameObject
			Vector3 sectionCopyPosition = slidingSectionInfo.sectionObject.transform.position;
			sectionCopyPosition.z -= GetOuterLevelSize();
			slidingSectionObjectCopy = CreateSectionObject(slidingSectionInfo, sectionCopyPosition);
			
			// set other variables required for sliding
			slidingSectionObjectCopyStartPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition.z += 1.0f;
			slidingPlayerStartPosition = player.transform.position;
			slidingPlayerEndPosition = player.transform.position;
			if (player.GetSectionIndices().x == x)
			{
				slidingPlayerEndPosition.z += direction;
			}

			// modify maze
			for (int z=GetOuterLevelSize()-2; z>=0; --z)
			{
				maze[x, z+1] = maze[x, z];
			}
			maze[x, 0] = slidingSectionInfo;
			
			// remember the section object that will be dropped
			slidingSectionObject = slidingSectionInfo.sectionObject;
			
			// set the new section object in the maze
			slidingSectionInfo.sectionObject = slidingSectionObjectCopy;
		}
		else // backward
		{
			// slidingSectionInfo: backmost section of the row
			slidingSectionInfo = maze[x, 0];
			
			// copy gameObject
			Vector3 sectionCopyPosition = slidingSectionInfo.sectionObject.transform.position;
			sectionCopyPosition.z += GetOuterLevelSize();
			slidingSectionObjectCopy = CreateSectionObject(slidingSectionInfo, sectionCopyPosition);
			
			// set other variables required for sliding
			slidingSectionObjectCopyStartPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition = sectionCopyPosition;
			slidingSectionObjectCopyEndPosition.z -= 1.0f;
			slidingPlayerStartPosition = player.transform.position;
			slidingPlayerEndPosition = player.transform.position;
			if (player.GetSectionIndices().x == x)
			{
				slidingPlayerEndPosition.z += direction;
			}
			
			// modify maze
			for (int z=1; z<GetOuterLevelSize(); ++z)
			{
				maze[x, z-1] = maze[x, z];
			}
			maze[x, GetOuterLevelSize()-1] = slidingSectionInfo;

			// remember the section object that will be dropped
			slidingSectionObject = slidingSectionInfo.sectionObject;
			
			// set the new section object in the maze
			slidingSectionInfo.sectionObject = slidingSectionObjectCopy;
		}
		
		ChangeState(State.SLIDING);
	}
	
	private bool IsSlidingLeftPossible()
	{
		bool slidingPossible = true;
		
		if (IsLeftSideAtEdge())
		{
			slidingPossible = false;
			//Debug.Log("cannot slide left if your left side is facing the edge of the level");
		}
		else if (!IsPlayerSectionMoveableLeftRight())
		{
			slidingPossible = false;
			//Debug.Log("cannot slide left because a locked section is blocking");			
		}
		
		return slidingPossible;
	}
	
	private bool IsSlidingRightPossible()
	{
		bool slidingPossible = true;
		
		if (IsRightSideAtEdge())
		{
			slidingPossible = false;
			//Debug.Log("cannot slide left if your right side is facing the edge of the level");
		}
		else if (!IsPlayerSectionMoveableLeftRight())
		{
			slidingPossible = false;
			//Debug.Log("cannot slide right because a locked section is blocking");			
		}
		
		return slidingPossible;
	}
	
	private bool IsSlidingForwardPossible()
	{
		bool slidingPossible = true;
		
		if (IsFacingEdge())
		{
			slidingPossible = false;			
			//Debug.Log("cannot slide forward if you're facing the edge of the level");
		}
		else if (!IsPlayerSectionMoveableForwardBackward())
		{
			slidingPossible = false;			
			//Debug.Log("cannot slide forward because a locked section is blocking");
		}
		
		return slidingPossible;
	}
	
	private bool IsSlidingBackwardPossible()
	{
		bool slidingPossible = true;
		
		if (IsBackingEdge())
		{
			slidingPossible = false;			
			//Debug.Log("cannot slide backward if you're facing the edge of the level");
		}
		else if (!IsPlayerSectionMoveableForwardBackward())
		{
			slidingPossible = false;			
			//Debug.Log("cannot slide backward because a locked section is bocking");
		}
		
		return slidingPossible;
	}
	
	private bool IsFacingEdge()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool facingEdge = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			facingEdge = (playerIndices.y >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "E")
		{
			facingEdge = (playerIndices.x >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "S")
		{
			facingEdge = (playerIndices.y <= 0);
		}
		else if (playerDirection == "W")
		{
			facingEdge = (playerIndices.x <= 0);
		}
		
		return facingEdge;
	}
	
	private bool IsBackingEdge()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool backingEdge = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			backingEdge = (playerIndices.y <= 0);
		}
		else if (playerDirection == "E")
		{
			backingEdge = (playerIndices.x <= 0);
		}
		else if (playerDirection == "S")
		{
			backingEdge = (playerIndices.y >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "W")
		{
			backingEdge = (playerIndices.x >= GetOuterLevelSize()-1);
		}
		
		return backingEdge;
	}
	
	private bool IsLeftSideAtEdge()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool leftSideAtEdge = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			leftSideAtEdge = (playerIndices.x <= 0);
		}
		else if (playerDirection == "E")
		{
			leftSideAtEdge = (playerIndices.y >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "S")
		{
			leftSideAtEdge = (playerIndices.x >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "W")
		{
			leftSideAtEdge = (playerIndices.y <= 0);
		}
		
		return leftSideAtEdge;		
	}
	
	private bool IsRightSideAtEdge()
	{
		Vector2 playerIndices = player.GetSectionIndices();
		bool rightSideAtEdge = false;
		
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			rightSideAtEdge = (playerIndices.x >= GetOuterLevelSize()-1);
		}
		else if (playerDirection == "E")
		{
			rightSideAtEdge = (playerIndices.y <= 0);
		}
		else if (playerDirection == "S")
		{
			rightSideAtEdge = (playerIndices.x <= 0);
		}
		else if (playerDirection == "W")
		{
			rightSideAtEdge = (playerIndices.y >= GetOuterLevelSize()-1);
		}
		
		return rightSideAtEdge;
	}
	
	private bool IsNextSectionMoveableLeftRight()
	{
		bool moveable = false;

		Vector2 playerIndices = player.GetSectionIndices();
		
		string playerDirection = player.GetDirection();
		
		if (playerDirection == "N")
		{
			if ((playerIndices.y + 1 < GetOuterLevelSize()) && (playerIndices.y %2 == 1))
			{
				moveable = true;
			}
		}
		else if (playerDirection == "E")
		{
			if ((playerIndices.x + 1 < GetOuterLevelSize()) && (playerIndices.x %2 == 1))
			{
				moveable = true;
			}
		}
		else if (playerDirection == "S")
		{
			if ((playerIndices.y - 1 >= 0) && (playerIndices.y %2 == 1))
			{
				moveable = true;
			}
		}
		else if (playerDirection == "W")
		{
			if ((playerIndices.x - 1 >= 0) && (playerIndices.x %2 == 1))
			{
				moveable = true;
			}
		}
		
		return moveable;
	}
	
	private bool IsPlayerSectionMoveableLeftRight()
	{
		bool moveable = false;
		
		string playerDirection = player.GetDirection();
		Vector2 playerIndices = player.GetSectionIndices();
		if (playerDirection == "N" || playerDirection == "S")
		{
			moveable = (playerIndices.y % 2 == 0);
		}
		else
		{
			moveable = (playerIndices.x % 2 == 0);
		}
		
		return moveable;
	}
	
	private bool IsPlayerSectionMoveableForwardBackward()
	{
		bool moveable = false;
		
		string playerDirection = player.GetDirection();
		Vector2 playerIndices = player.GetSectionIndices();
		if (playerDirection == "N" || playerDirection == "S")
		{
			moveable = (playerIndices.x % 2 == 0);
		}
		else
		{
			moveable = (playerIndices.y % 2 == 0);
		}
		
		return moveable;
	}
	
	private void ChangeState(State newState)
	{
		// exit old state
		if (state == State.MOVING)
		{
		}
		else if (state == State.SLIDING)
		{
			slidingSectionObjectCopy.rigidbody.isKinematic = false;
		}
		else if (state == State.SLIDING_ADJUSTING)
		{
			// set all kinematic false
			foreach (SectionInfo sectionInfo in maze)
			{
				sectionInfo.sectionObject.rigidbody.isKinematic = false;
			}
			
			// destroy slidingSectionObject (the object that fell outside the game)
			Destroy(slidingSectionObject);
			slidingSectionObject = null;
			
			adjustingStartPositions = null; 
			adjustingStartQuaternions = null;			
		}
		
		// init new state
		if (newState == State.MOVING)
		{
			ignoringInput = true;
		}
		else if (newState == State.SLIDING)
		{
			slidingTimeLeft = SLIDING_TIME;
			slidingSectionObjectCopy.rigidbody.isKinematic = true;
		}
		else if (newState == State.SLIDING_ADJUSTING)
		{
			adjustingTimeLeft = ADJUSTING_TIME;
						
			// set all kinematic true
			foreach (SectionInfo sectionInfo in maze)
			{
				sectionInfo.sectionObject.rigidbody.isKinematic = true;
			}
			
			adjustingStartPositions = new Vector3[GetOuterLevelSize(),GetOuterLevelSize()];
			adjustingStartQuaternions = new Quaternion[GetOuterLevelSize(),GetOuterLevelSize()];
			
			for (int x = 0; x < GetOuterLevelSize(); ++x)
			{
				for (int z = 0; z < GetOuterLevelSize(); ++z)
				{
					adjustingStartPositions[x,z] = maze[x,z].sectionObject.transform.position;
					adjustingStartQuaternions[x,z] = maze[x,z].sectionObject.transform.rotation;
				}
			}
		}
		
		// set state
		state = newState;
	}
	
	private GameObject CreateSectionObject(SectionInfo sectionInfo, Vector3 position)
	{
		GameObject sectionObject = (GameObject)Instantiate( sections[(int)sectionInfo.sectionType],
		                                                    position,
		                                                    sectionRotationQuaternions[(int)sectionInfo.rotation] );
		return sectionObject;
	}

	private Vector3 SectionIndexToCoordinate(Vector2 sectionIndex)
	{
		Vector3 coordinate = new Vector3(-(GetOuterLevelSize() / 2.0f) + 0.5f, -0.07f, -(GetOuterLevelSize() / 2.0f) + 0.5f);		 
		coordinate.x += sectionIndex.x;
		coordinate.z += sectionIndex.y;
		
		return coordinate;
	}
	
	private string GetArrowCharacter()
	{
		string arrowCharacter = " ";
		string playerDirection = player.GetDirection();
		if (playerDirection == "N")
		{
			arrowCharacter = "1";
		}
		else if (playerDirection == "E")
		{
			arrowCharacter = ">";
		}
		else if (playerDirection == "S")
		{
			arrowCharacter = "v";
		}
		else if (playerDirection == "W")
		{
			arrowCharacter = "<";
		}
		
		return arrowCharacter;
	}
	
	private int GetLevelEndHealthIncrement()
	{
		return (100 - health)/2;
	}
	
	private float GetPercentageExplored()
	{	
		int numVisited = 0;
		foreach(SectionInfo sectionInfo in maze)
		{
			if (sectionInfo.visited)
			{
				++numVisited;
			}
		}
		return (((numVisited * 1.0f)/(GetOuterLevelSize()*GetOuterLevelSize()))*100.0f);
	}
	
	private void CreateOtherLevelObjects()
	{
		int positionsAvailable = (innerLevelSize*innerLevelSize)-1;
		// 50% money + treasure
		int numMoneyObjects = (int)(positionsAvailable * 0.5f * 0.75f);
		int numTreasureObjects = (int)(positionsAvailable * 0.5f * 0.25f);
		// 25% medikits
		int numMedikitObjects = (int)(positionsAvailable * 0.25f);
		
		int x = 0;
		int y = 0;
		
		bool[,] occupied = new bool[innerLevelSize,innerLevelSize];
		for (x=0; x<innerLevelSize; ++x)
		{
			for (y = 0; y<innerLevelSize; ++y)
			{
				occupied[x,y] = false;
			}
		}
		
		x = 0;
		y = 0;
	
		moneyObjects = new ArrayList(numMoneyObjects);
		for (int i=0; i<numMoneyObjects; ++i)
		{
			x = Random.Range(0, innerLevelSize);
			y = Random.Range(0, innerLevelSize);
			while ( (targetPosition == new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y)))
				    || occupied[x,y] )
			{
				x = Random.Range(0, innerLevelSize);
				y = Random.Range(0, innerLevelSize);
			}
			
			moneyObjects.Add((LevelObject)((GameObject)Instantiate( moneyPrefab,
				                                                    SectionIndexToCoordinate(new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y))),
				                                                    Quaternion.identity )).GetComponent(typeof(LevelObject)) );
			
			occupied[x,y] = true;
		}
		
		treasureObjects = new ArrayList(numTreasureObjects);
		for (int i=0; i<numTreasureObjects; ++i)
		{
			x = Random.Range(0, innerLevelSize);
			y = Random.Range(0, innerLevelSize);
			while ( (targetPosition == new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y)))
				    || occupied[x,y] )
			{
				x = Random.Range(0, innerLevelSize);
				y = Random.Range(0, innerLevelSize);
			}
			
			treasureObjects.Add((LevelObject)((GameObject)Instantiate( treasurePrefab,
				                                                       SectionIndexToCoordinate(new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y))),
				                                                       Quaternion.identity )).GetComponent(typeof(LevelObject)) );
			
			occupied[x,y] = true;
		}
		
		medikitObjects = new ArrayList(numMedikitObjects);
		for (int i=0; i<numMedikitObjects; ++i)
		{
			x = Random.Range(0, innerLevelSize);
			y = Random.Range(0, innerLevelSize);
			while ( (targetPosition == new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y)))
				    || occupied[x,y] )
			{
				x = Random.Range(0, innerLevelSize);
				y = Random.Range(0, innerLevelSize);
			}
			
			medikitObjects.Add((LevelObject)((GameObject)Instantiate( medikitPrefab,
	                                                                  SectionIndexToCoordinate(new Vector2(InnerCoordinateToOuter(x), InnerCoordinateToOuter(y))),
	                                                                  Quaternion.identity )).GetComponent(typeof(LevelObject)) );

			occupied[x,y] = true;
		}
	}
	
	private bool ShowMessageBox(string message)
	{
		bool continueButtonPressed = false;
		
		// show box
		GUI.DrawTexture(new Rect((Screen.width - 465.0f*0.90f)/2.0f, (Screen.height - 457.0f*0.90f)/2.0f, 465.0f*0.90f, 457.0f*0.90f), hudTex_messageBox);
		
		TextAnchor alignmentBackup = GUI.skin.label.alignment;
		{
			GUI.skin.label.alignment = TextAnchor.MiddleCenter;
			GUI.Label(MESSAGE_BOX_TEXT_RECT, message );
		}
		GUI.skin.label.alignment = alignmentBackup;
		
		if ((CONTINUE_BUTTON_RECT.Contains(Util.GetInvertedMousePosition()) && Input.GetMouseButtonDown(0)) || Input.GetKeyDown(KeyCode.Space))
		{
			continueButtonPressed = true;
		}
		
		return continueButtonPressed;
	}
	
	private void PlayRandomSlideSound()
	{
		int r = Random.Range(0, 3);
		if (r == 0)
		{
			audio.PlayOneShot(sfxSliding1);
		}
		else if (r == 1)
		{
			audio.PlayOneShot(sfxSliding2);
		}
		else if (r == 2)
		{
			audio.PlayOneShot(sfxSliding3);
		}
	}
}
