using UnityEngine;
using System.Collections;

public class Player : LevelObject
{
	private bool moveForwardPossible = true;
	private bool moveBackwardPossible = true;
	private bool moveForwardButtonUsed = false;
	private bool moveLeftButtonUsed = false;
	private bool moveRightButtonUsed = false;
	private bool moveBackwardButtonUsed = false;
	private Texture hudTex_controllerForward = null;
	private Texture hudTex_controllerLeft = null;
	private Texture hudTex_controllerRight = null;
	private AudioClip sfxSteps1;
	private AudioClip sfxSteps2;
	private AudioClip sfxSteps3;
	private AudioClip sfxSteps4;
	
	//TUNING
	// see drawing for explanation for coordinates
	private Rect LEFT_BUTTON_RECT = new Rect(198.0f, 379.0f, 61.0f, 62.0f);
	private Rect RIGHT_BUTTON_RECT = new Rect(342.0f, 379.0f, 60.0f, 57.0f);
	private Rect FORWARD_BUTTON_RECT = new Rect(266.0f, 379.0f, 67.0f, 64.0f);

	private void Awake()
	{
		hudTex_controllerForward = (Texture)Resources.Load("Textures/controller_forward");
		hudTex_controllerLeft = (Texture)Resources.Load("Textures/controller_left");
		hudTex_controllerRight = (Texture)Resources.Load("Textures/controller_right");
		
		sfxSteps1 = (AudioClip)Resources.Load("Sfx/steps1");
		sfxSteps2 = (AudioClip)Resources.Load("Sfx/steps2");
		sfxSteps3 = (AudioClip)Resources.Load("Sfx/steps3");
		sfxSteps4 = (AudioClip)Resources.Load("Sfx/steps4");
	}
	
	public void Update()
	{
		if (Game.GetState() == Game.State.MOVING)
		{
			if ((LEFT_BUTTON_RECT.Contains(Util.GetInvertedMousePosition()) && Input.GetMouseButtonDown(0)) || Input.GetKeyDown(KeyCode.LeftArrow))
			{
				if (!Game.ignoringInput)
				{
					moveLeftButtonUsed = true;
				}
			}
		
			if ((RIGHT_BUTTON_RECT.Contains(Util.GetInvertedMousePosition()) && Input.GetMouseButtonDown(0)) || Input.GetKeyDown(KeyCode.RightArrow))
			{
				if (!Game.ignoringInput)
				{
					moveRightButtonUsed = true;
				}
			}
		
			if (moveForwardPossible)
			{
				if ((FORWARD_BUTTON_RECT.Contains(Util.GetInvertedMousePosition()) && Input.GetMouseButtonDown(0)) || Input.GetKeyDown(KeyCode.UpArrow))
				{
					if (!Game.ignoringInput)
					{
						moveForwardButtonUsed = true;
					}
				}
			}
		
			if (moveBackwardPossible && Input.GetKeyDown(KeyCode.DownArrow))
			{
				if (!Game.ignoringInput)
				{
					moveBackwardButtonUsed = true;
				}
			}
			
			if (moveForwardButtonUsed)
			{
				transform.position += transform.forward;
				Game.health -= Game.HEALTH_DECREASE_WALK;
				PlayRandomMoveSound();
			}
			else if (moveLeftButtonUsed)
			{
				transform.RotateAround(Vector3.up, Mathf.Deg2Rad * -90.0f);
			}
			else if (moveRightButtonUsed)
			{
				transform.RotateAround(Vector3.up, Mathf.Deg2Rad * 90.0f);
			}
			else if (moveBackwardButtonUsed)
			{
				transform.position -= transform.forward;
				Game.health -= Game.HEALTH_DECREASE_WALK;
				PlayRandomMoveSound();
			}
		}
	}
	
	public void LateUpdate()
	{
		moveForwardButtonUsed = false;
		moveLeftButtonUsed = false;
		moveRightButtonUsed = false;
		moveBackwardButtonUsed = false;
	}
	
	public void OnGUI()
	{
		if (Game.GetState() == Game.State.MOVING)
		{
			GUI.DrawTexture(LEFT_BUTTON_RECT, hudTex_controllerLeft);							
			GUI.DrawTexture(RIGHT_BUTTON_RECT, hudTex_controllerRight);			
			
			if (moveForwardPossible) 
			{
				GUI.DrawTexture(FORWARD_BUTTON_RECT, hudTex_controllerForward);
			}
		}	
	}
	
	public void SetMoveForwardPossible(bool moveForwardPossible)
	{
		this.moveForwardPossible = moveForwardPossible;
	}
	
	public void SetMoveBackwardPossible(bool moveBackwardPossible)
	{
		this.moveBackwardPossible = moveBackwardPossible;
	}
			


	public string GetDirection()
	{
		string direction = "";
				
		if (MathExt.Equals(transform.forward.x, 0.0f) && MathExt.Equals(transform.forward.z, 1.0f))
		{
			direction = "N";
		}
		else if (MathExt.Equals(transform.forward.x, 1.0f) && MathExt.Equals(transform.forward.z, 0.0f))
		{
			direction = "E";
		}
		else if (MathExt.Equals(transform.forward.x, 0.0f) && MathExt.Equals(transform.forward.z, -1.0f))
		{
			direction = "S";
		}
		else if (MathExt.Equals(transform.forward.x, -1.0f) && MathExt.Equals(transform.forward.z, 0.0f))
		{
			direction = "W";
		}
		
		return direction;
	}
	
	private void PlayRandomMoveSound()
	{
		int r = Random.Range(0, 4);
		if (r == 0)
		{
			audio.PlayOneShot(sfxSteps1);
		}
		else if (r == 1)
		{
			audio.PlayOneShot(sfxSteps2);
		}
		else if (r == 2)
		{
			audio.PlayOneShot(sfxSteps3);
		}
		else if (r == 3)
		{
			audio.PlayOneShot(sfxSteps4);
		}
	}
}
