using UnityEngine;
using System.Collections.Generic;

public class Section : MonoBehaviour {
	
	private List<GameObject> floorPlanes = new List<GameObject>(8);
	private List<GameObject> innerWallPlanes = new List<GameObject>(8);
	private List<GameObject> outerWallPlanes = new List<GameObject>(8);
	
	
	private Texture fixedFloorTexture = null;
	
	private Texture floorTexture = null;
	private Texture floorTexture1 = null;
	private Texture floorTexture2 = null;
	
	private Texture innerWallTexture = null;
	private Texture innerWallTexture1 = null;
	private Texture innerWallTexture2 = null;
	
	private Texture outerWallTexture = null;
	private Texture outerWallTexture1 = null;
	private Texture outerWallTexture2 = null;
	
	private const float SPECIAL_TEXTURE_CHANCE = 0.1f;
	
	public void Awake ()
	{
		// find objects that need to be textured
		foreach (Transform child in transform)
		{
			if (child.name == "TilePlane") // floor
			{
				foreach (Transform grandChild in child)
				{
					floorPlanes.Add(grandChild.gameObject);
				}
			}
			else if (child.name == "DoubleSidedPlane") // walls
			{
				foreach (Transform grandChild in child)
				{
					if (grandChild.name == "Plane_inner")
					{
						innerWallPlanes.Add(grandChild.gameObject);
					}
					else
					{
						outerWallPlanes.Add(grandChild.gameObject);
					}

				}
			}
		}
		
		// get the textures
		fixedFloorTexture = (Texture)Resources.Load("Textures/fixedFloor");
	
		floorTexture = (Texture)Resources.Load("Textures/floor");
		floorTexture1 = (Texture)Resources.Load("Textures/floor_1");
		floorTexture2 = (Texture)Resources.Load("Textures/floor_2");
		
		innerWallTexture = (Texture)Resources.Load("Textures/innerWall");
		innerWallTexture1 = (Texture)Resources.Load("Textures/innerWall_1");
		innerWallTexture2 = (Texture)Resources.Load("Textures/innerWall_2");
		
		outerWallTexture = (Texture)Resources.Load("Textures/outerWall");
		outerWallTexture1 = (Texture)Resources.Load("Textures/outerWall_1");
		outerWallTexture2 = (Texture)Resources.Load("Textures/outerWall_2");
		
		// set anisotropy levels
		fixedFloorTexture.anisoLevel = 4;
		
		floorTexture.anisoLevel = 4;
		floorTexture1.anisoLevel = 4;
		floorTexture2.anisoLevel = 4;
		
		innerWallTexture.anisoLevel = 4;
		innerWallTexture1.anisoLevel = 4;
		innerWallTexture2.anisoLevel = 4;
		
		outerWallTexture.anisoLevel = 4;
		outerWallTexture1.anisoLevel = 4;
		outerWallTexture2.anisoLevel = 4;
		
		// set wall textures
		{
			// inner
			foreach(GameObject plane in innerWallPlanes)
			{
				if (Random.Range(0.0f, 1.0f) > SPECIAL_TEXTURE_CHANCE)
				{
					plane.renderer.material.mainTexture = innerWallTexture;
				}
				else
				{
					if (Random.Range(0.0f, 1.0f) < 0.5f)
					{
						plane.renderer.material.mainTexture = innerWallTexture1;				
					}
					else
					{
						plane.renderer.material.mainTexture = innerWallTexture2;				
					}
				}
			}
			
			// outer
			foreach(GameObject plane in outerWallPlanes)
			{
				if (Random.Range(0.0f, 1.0f) > SPECIAL_TEXTURE_CHANCE)
				{
					plane.renderer.material.mainTexture = outerWallTexture;
				}
				else
				{
					if (Random.Range(0.0f, 1.0f) < 0.5f)
					{
						plane.renderer.material.mainTexture = outerWallTexture1;				
					}
					else
					{
						plane.renderer.material.mainTexture = outerWallTexture2;				
					}
				}
			}
		}
			
		// also sets the floor texture
		AllowMove(true);
	}
	
	public void Update ()
	{
	
	}
	
	public void AllowMove(bool moveAllowed)
	{
		if (moveAllowed)
		{		
			foreach(GameObject plane in floorPlanes)
			{
				if (Random.Range(0.0f, 1.0f) > SPECIAL_TEXTURE_CHANCE)
				{
					plane.renderer.material.mainTexture = floorTexture;
				}
				else
				{
					if (Random.Range(0.0f, 1.0f) < 0.5f)
					{
						plane.renderer.material.mainTexture = floorTexture1;				
					}
					else
					{
						plane.renderer.material.mainTexture = floorTexture2;				
					}
				}
			}
		}
		else
		{
			foreach(GameObject plane in floorPlanes)
			{
				plane.renderer.material.mainTexture = fixedFloorTexture;
			}				
		}
	}
}
